/*
 * Copyright (c) MCST, 2021
 *
 * Test Name: ioctl_userfaultfd
 *
 * Test is intended to verify userfaultfd ioctl interface for
 * UFFDIO_REGISTER_MODE_MISSING mode.
 * The followng ioctl cmds are verified:
 *    UFFDIO_API
 *    UFFDIO_REGISTER
 *    UFFDIO_UNREGISTER
 *    UFFDIO_COPY
 *    UFFDIO_ZEROPAGE
 */

#ifdef __ptr128__

#define _GNU_SOURCE

#include <sys/types.h>
#include <stdio.h>
#include <pthread.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <poll.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <linux/userfaultfd.h>

#include "test.h"
#include "safe_macros.h"
#include "lapi/syscalls.h"

#define NUM_PAGES 2

char *TCID = "ioctl_userfaultfd";

static int testno;

static int page_size;
static int zeropage_ok, thread_id_ok;
static int zeropage_used;
static long fd;

static void userfaultfd_ioctl(void);

struct test_case_t {
    void (*setup) (void);
    void (*test_all) (void);
    void (*cleanup) (void);
    char *desc;
} test_data[] = {
    { NULL, userfaultfd_ioctl, NULL,
            "userfaultfd ioctl()"}
,};

int TEST_TOTAL = sizeof(test_data) / sizeof(test_data[0]);

int main(int argc, char *argv[])
{
    int lc;

    tst_parse_opts(argc, argv, NULL, NULL);
    TEST_PAUSE;

    for (lc = 0; TEST_LOOPING(lc); ++lc) {
        tst_count = 0;
        for (testno = 0; testno < TEST_TOTAL; ++testno) {
            if (test_data[testno].setup != NULL)
                test_data[testno].setup();
            if (test_data[testno].test_all != NULL)
                test_data[testno].test_all();
            if (test_data[testno].cleanup != NULL)
                test_data[testno].cleanup();
        }
    }

    tst_exit();
}

static void * fault_handler_thread(void *arg)
{
    struct uffd_msg msg;
    int event = 0;
    char *page = NULL;
    pid_t pid = getpid();

    /* A page will be copied into the faulting region
       by UFFDIO_COPY ioctl. */
    page = mmap(NULL, page_size, PROT_READ | PROT_WRITE,
                MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
    if (page == MAP_FAILED)
         tst_brkm(TBROK, NULL, "mmap failed");

    /* Handling events. */
    for (;;) {
         struct pollfd pollfd;
         pollfd.fd = fd;
         pollfd.events = POLLIN;
         TEST(poll(&pollfd, 1, -1));
         if (TEST_RETURN == -1)
              tst_brkm(TBROK, NULL, "poll(): errno=%d", TEST_ERRNO);

         TEST(read(fd, &msg, sizeof(msg)));
         if (TEST_RETURN == 0)
              tst_brkm(TBROK, NULL, "read(): unexpected EOF");
         if (TEST_RETURN == -1)
              tst_brkm(TBROK, NULL, "read(): errno=%d", TEST_ERRNO);

         if (msg.event != UFFD_EVENT_PAGEFAULT)
              tst_brkm(TBROK, NULL, "Unexpected event: %d", msg.event);

         if (thread_id_ok) {
              if (msg.arg.pagefault.feat.ptid == pid)
                   tst_resm(TPASS, "Received ptid as expected");
              else 
                   tst_resm(TFAIL, "Received tid(%d) is not equal to a process ID(%d)",
                            msg.arg.pagefault.feat.ptid, pid);
         }

         if (zeropage_ok && event % 2) {
              /* UFFDIO_ZEROPAGE */
              struct uffdio_zeropage uffdio_zeropage;
              zeropage_used = 1;
              uffdio_zeropage.range.start = (unsigned long) msg.arg.pagefault.address &
                                                  ~(page_size - 1);
              uffdio_zeropage.range.len = page_size;
              uffdio_zeropage.mode = 0;
              uffdio_zeropage.zeropage = 0;
              TEST(ioctl(fd, UFFDIO_ZEROPAGE, &uffdio_zeropage));
              if (TEST_RETURN == -1)
                  tst_brkm(TBROK, NULL, "ioctl(UFFDIO_ZEROPAGE): errno=%d",
                           TEST_ERRNO);

              if (uffdio_zeropage.zeropage == page_size)
                     tst_resm(TPASS, "ioctl(UFFDIO_ZEROPAGE) returned %d"
                              " as expected", (int)uffdio_zeropage.zeropage);
              else
                     tst_resm(TFAIL, "ioctl(UFFDIO_ZEROPAGE) returned %d",
                              (int)uffdio_zeropage.zeropage);
         } else {
              /* UFFDIO_COPY */
              struct uffdio_copy uffdio_copy;
              zeropage_used = 0;
              /* Copy the 'colored' page into the faulting region. */
              memset(page, 'A', page_size);
              uffdio_copy.src = (unsigned long) page;
              uffdio_copy.dst = (unsigned long) msg.arg.pagefault.address &
                                                  ~(page_size - 1);
              uffdio_copy.len = page_size;
              uffdio_copy.mode = 0;
              uffdio_copy.copy = 0;
              TEST(ioctl(fd, UFFDIO_COPY, &uffdio_copy));
              if (TEST_RETURN == -1)
                  tst_brkm(TBROK, NULL, "ioctl(UFFDIO_COPY): errno=%d",
                           TEST_ERRNO);

              if (uffdio_copy.copy == page_size)
                     tst_resm(TPASS, "ioctl(UFFDIO_COPY) returned %d"
                              " as expected", (int)uffdio_copy.copy);
              else
                     tst_resm(TFAIL, "ioctl(UFFDIO_COPY) returned %d",
                              (int)uffdio_copy.copy);
         }
         event++;
    }
}

static void userfaultfd_ioctl(void)
{
    char *addr;         /* Start of region handled by userfaultfd */
    uint64_t len;       /* Length of region handled by userfaultfd */
    pthread_t thr;
    struct uffdio_api uffdio_api;
    struct uffdio_register uffdio_register;
    struct uffdio_range uffdio_range;
    int loc = 0x10;

    page_size = sysconf(_SC_PAGE_SIZE);
    if (page_size == -1)
          tst_brkm(TBROK, NULL, "sysconf(_SC_PAGE_SIZE) failed");

    len = NUM_PAGES * page_size;

    fd = syscall(__NR_userfaultfd, O_CLOEXEC | O_NONBLOCK);
    if (fd == -1)
          tst_brkm(TBROK, NULL, "syscall(__NR_userfaultfd) failed");

    uffdio_api.api = UFFD_API;
    uffdio_api.features = UFFD_FEATURE_THREAD_ID;
    TEST(ioctl(fd, UFFDIO_API, &uffdio_api));
    if (TEST_RETURN < 0) {
          tst_brkm(TBROK, NULL, "ioctl(UFFDIO_API): errno=%d",
                   TEST_ERRNO);
    } else {
          tst_resm(TPASS, "ioctl(UFFDIO_API)");
    }

    if (uffdio_api.features & UFFD_FEATURE_THREAD_ID) {
          thread_id_ok = 1;
          tst_resm(TPASS, "UFFD_FEATURE_THREAD_ID supported");
    } else {
          tst_resm(TFAIL, "UFFD_FEATURE_THREAD_ID unsupported but expected!");
    }

    addr = mmap(NULL, len, PROT_READ | PROT_WRITE,
                       MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
    if (addr == MAP_FAILED)
          tst_brkm(TBROK, NULL, "mmap failed");

    uffdio_register.range.start = (unsigned long) addr;
    uffdio_register.range.len = len;
    uffdio_register.mode = UFFDIO_REGISTER_MODE_MISSING;
    TEST(ioctl(fd, UFFDIO_REGISTER, &uffdio_register));
    if (TEST_RETURN < 0) {
          tst_brkm(TBROK, NULL, "ioctl(UFFDIO_REGISTER): errno=%d",
                   TEST_ERRNO);
    } else {
          tst_resm(TPASS, "ioctl(UFFDIO_REGISTER)");
    }

    if (uffdio_register.ioctls == UFFD_API_RANGE_IOCTLS) {
          zeropage_ok = 1;
          tst_resm(TINFO, "UFFDIO_ZEROPAGE supported");
    } else {
          tst_resm(TINFO, "UFFDIO_ZEROPAGE not supported");
    }

    /* A thread will process the userfaultfd events. */
    TEST(pthread_create(&thr, NULL, fault_handler_thread, NULL));
    if (TEST_RETURN != 0) {
          errno = TEST_RETURN;
          tst_brkm(TBROK, NULL, "pthread_create(): errno=%d", errno);
    }

    /* Let's touch memory in the mapping. This will trigger userfaultfd
       events for all pages in the region. */
    while ((uint64_t)loc < len) {
          char c = addr[loc];
          if (c == 'A' && !zeropage_used)
               tst_resm(TPASS, "Page fault handled by UFFDIO_COPY as expected");
          else if (c == '\0' && zeropage_used)
               tst_resm(TPASS, "Page fault handled by UFFDIO_ZEROPAGE as expected");
          else
               tst_resm(TFAIL, "Unexpected page 'color'!");
          loc += page_size;
    }

    uffdio_range.start = (unsigned long) addr;
    uffdio_range.len = len;
    TEST(ioctl(fd, UFFDIO_UNREGISTER, &uffdio_range));
    if (TEST_RETURN == -1)
          tst_brkm(TBROK, NULL, "ioctl(UFFDIO_UNREGISTER): errno=%d",
                   TEST_ERRNO);
    else
          tst_resm(TPASS, "ioctl(UFFDIO_UNREGISTER)");
}

#else /* __ptr128__ */

/* NB The test makes no sense in non-protected environment */

int main(void)
{
	return 32; /* test is to be skipped for this platform  */
}

#endif /* __ptr128__ */
